class DebugExecutor:
    def __init__(self, max_workers=None):
        pass

    def submit(self, fn, *args, **kwargs):
        return DebugFuture(fn, *args, **kwargs)

class DebugFuture:
    def __init__(self, fn, *args, **kwargs):
        self.fn = fn
        self.args = args
        self.kwargs = kwargs
        self._condition = DebugCondition()
        self._state = 'FINISHED'
        self._waiters = []

    def result(self):
        return self.fn(*self.args, **self.kwargs)

class DebugCondition:
    def __init__(self):
        self.acquire = lambda: None
        self.release = lambda: None

    def __enter__(self):
        pass

    def __exit__(self, *args):
        pass




import importlib, re

def http_to_https(url):
    if url.startswith("http://"):
        return url.replace("http://", "https://", 1)
    return url

def cdn_url_to_http(url):
    if url.startswith("//"):
        return "http://" + url.lstrip("//")
    return url

# TODO: rename to work with methods too
def import_class(s):
    try:
        module_string, klass_string = s.rsplit('.', 1)
    except ValueError:
        raise ValueError('could not parse {}. provide the full path to class'.format(s))
    module = importlib.import_module(module_string)
    klass = getattr(module, klass_string)
    return klass

def get_only_first_or_none(lst):
    if len(lst) > 1: raise ValueError("too many values")
    return lst[0] if len(lst) else None

def convert_dict_to_camel_case(d): # pragma: no cover
    from boltons.iterutils import remap

    re1 = re.compile(r'(.)([A-Z][a-z]+)')
    re2 = re.compile(r'([a-z0-9])([A-Z])')

    def convert_str_to_camel_case(s):
        s1 = re.sub(re1, r'\1_\2', s)
        return re.sub(re2, r'\1_\2', s1).lower()

    def visit(path, key, value):
        new_key = convert_str_to_camel_case(key) if isinstance(key, str) else key
        return new_key, value

    return remap(d, visit=visit)

async def aenumerate(aiterable):
    i = 0
    async for x in aiterable:
        yield i, x
        i += 1

# https://stackoverflow.com/questions/42378566/python-3-6-async-version-of-islice
async def aislice(aiterable, *args):
    s = slice(*args)
    it = iter(range(s.start or 0, s.stop or sys.maxsize, s.step or 1))
    try:
        nexti = next(it)
    except StopIteration:
        return
    async for i, element in aenumerate(aiterable):
        if i == nexti:
            yield element
            try:
                nexti = next(it)
            except StopIteration:
                return
import lxml

def debug_doc(fn):
    def inner(doc):
        try:
            return fn(doc)
        except Exception as e:
            with open('test.html', 'wb') as f:
                f.write(lxml.html.tostring(doc))
            raise e
    return inner



import logging

logger = logging.getLogger(__name__)

def requests_get(url, timeout=10):
    logger.debug("downloading %s" % url)
    import requests
    return requests.get(url, timeout=timeout).text

async def asyncio_get(url, timeout=10):
    logger.debug("downloading %s" % url)
    import aiohttp
    import async_timeout
    async with aiohttp.ClientSession() as session:
        with async_timeout.timeout(timeout):
            async with session.get(url) as r:
                return await r.text()

# def asyncio_queue():
#     from asyncio import Queue
#     return Queue

# async def tornado_get():
#     from tornado import httpclient
#     # http_client = httpclient.AsyncHTTPClient(max_clients=CONCURRENCY)
#     # await http_client.fetch(url, connect_timeout=5, request_timeout=45)
#     pass

from craigslist.io.debug_executor import DebugExecutor



import json
import logging
from collections import namedtuple
from datetime import datetime, timezone
from craigslist.utils import cdn_url_to_http
from craigslist.exceptions import CraigslistException

logger = logging.getLogger(__name__)

JSONSearchPost = namedtuple('JSONSearchPost', [
    'id',
    'title',
    'url',
    'category_id',
    'thumbnail',
    'longitude',
    'latitude',
    'date',
    'price',
    'bedrooms'])

JSONSearchCluster = namedtuple('JSONSearchCluster', [
    'id',
    'url',
    'longitude',
    'latitude',
    'posting_ids',
    'num_posts',
    'date'])

def parse_cluster_url_output(body):
    try:
        items, meta = json.loads(body)
    except ValueError as e: # pragma: no cover
        raise CraigslistException(
            "could not find items and meta "
            "in json response body: '{}'".format(body))
    try:
        baseurl = cdn_url_to_http(meta['baseurl'])
    except KeyError as e: # pragma: no cover
        raise CraigslistException(
            "could not find baseurl in meta: '{}' "
            "with body:'{}'. probably empty response. "
            "is your query too specific?".format(meta, body))
        raise
    posts = [parse_post(x)
        for x in items if not x.get('GeoCluster')]

    clusters = [parse_cluster(x, baseurl)
        for x in items if x.get('GeoCluster')]
    return posts, clusters

def parse_cluster(cluster, baseurl):
    return JSONSearchCluster(**{
        'id': int(cluster['GeoCluster']),
        'url': baseurl + cluster['url'],
        'longitude': cluster['Longitude'],
        'latitude': cluster['Latitude'],
        'num_posts': cluster['NumPosts'],
        'posting_ids': [int(x) for x in cluster['PostingID'].split(',')],
        'date': datetime.fromtimestamp(
            float(cluster['PostedDate']), timezone.utc).isoformat()
    })

def parse_post(post):
    return JSONSearchPost(**{
        'id': int(post['PostingID']),
        'title': post['PostingTitle'],
        'url': cdn_url_to_http(post['PostingURL']),
        'category_id': post['CategoryID'],
        'thumbnail': post.get('ImageThumb'),
        'longitude': post['Longitude'],
        'latitude': post['Latitude'],
        'date': datetime.fromtimestamp(
            float(post['PostedDate']), timezone.utc).isoformat(),
        'price': post['Ask'],
        'bedrooms': post.get('Bedrooms'),
    })

import concurrent.futures
from craigslist._search import get_query_url
from craigslist.post import process_post_url

# http://stackoverflow.com/questions/1747963/multiprocessing-pool-inside-process-time-out/1748335#1748335
# You can't pass Pool objects between processes.

def process_cluster_url(url, get):
    logger.debug("downloading %s" % url)
    body = get(url)
    return parse_cluster_url_output(body)

def jsonsearch(
    area,
    category,
    sort,
    cache,
    cachedir,
    executor,
    get,
    as_completed=concurrent.futures.as_completed,
    **kwargs):

    def process_clusters(clusters, executor):
        futures = (executor.submit(
            process_cluster_url, cluster.url, get) for cluster in clusters)
        try:
            for future in as_completed(futures):
                posts, clusters = future.result()
                yield from posts
                process_clusters(clusters, executor)
        except KeyboardInterrupt:  # pragma: no cover
            for future in futures:
                future.cancel()

    url = get_query_url(area, category, "jsonsearch", sort=sort, map=1, **kwargs)
    posts, clusters = process_cluster_url(url, get)
    yield from posts
    yield from process_clusters(clusters, executor)

import asyncio

async def process_cluster_url_async(url, get):
    logger.debug("downloading %s" % url)
    body = await get(url)
    return parse_cluster_url_output(body)

async def jsonsearch_async(
    area,
    category,
    sort,
    cache,
    cachedir,
    get,
    as_completed=asyncio.as_completed,
    **kwargs):

    async def process_clusters(clusters):
        futures = [process_cluster_url_async(cluster.url, get) for cluster in clusters]
        try:
            for future in as_completed(futures):
                posts, clusters = await future
                for post in posts:
                    yield post
                async for post in process_clusters(clusters):
                    yield post
        except KeyboardInterrupt: # pragma: no cover
            for future in futures:
                future.cancel()

    url = get_query_url(area, category, "jsonsearch", sort=sort, map=1, **kwargs)
    posts, clusters = await process_cluster_url_async(url, get)

    for post in posts:
        yield post
    async for post in process_clusters(clusters):
        yield post



import os
import functools
from urllib.parse import urlencode
from craigslist.utils import cdn_url_to_http, import_class
from craigslist.data import get_areas, get_categories
from craigslist.io import requests_get, asyncio_get
from craigslist.post import get_posts, get_posts_async
from craigslist.exceptions import (
    CraigslistException, CraigslistValueError)

@functools.lru_cache(maxsize=None)
def get_url_base(area):
    areas = get_areas()
    if area not in areas:
        raise CraigslistValueError("unknown area {}".format(area))
    tld = areas[area]['tld']
    return "https://{}.craigslist.{}".format(area, tld)

def get_query_url(area, category, type_, offset=0, sort="date", **kwargs):
    categories = get_categories()
    if category not in categories:
        raise CraigslistValueError("unknown category {}".format(category))
    params = {"s": offset, "sort": sort, **kwargs}
    params_expanded = []
    for k,v in params.items():
        if v is None: continue
        if isinstance(v, list):
            for x in v:
                params_expanded.append((k, x))
        else:
            params_expanded.append((k,v))
    url = get_url_base(area) + "/{type}/{category}?{params}".format(
        type=type_, category=category, params=urlencode(params_expanded))
    return url

def make_executor(executor_class, max_workers=None):
    if isinstance(executor_class, str):
        executor_class = import_class(executor_class)
    return executor_class(max_workers=max_workers)

from craigslist._search.jsonsearch import jsonsearch
from craigslist._search.regularsearch import regularsearch

def search(
    area,
    category,
    type_="jsonsearch",
    get_detailed_posts=False,
    cache=True,
    cachedir=os.path.expanduser('~'),
    executor=None,
    executor_class='concurrent.futures.ThreadPoolExecutor',
    max_workers=None,
    get=requests_get,
    **kwargs):

    def extract_post_urls(posts):
        yield from (post.url for post in posts)

    executor = executor or make_executor(executor_class, max_workers)

    search_funcs = {
        "jsonsearch": jsonsearch,
        "regularsearch": regularsearch}

    try:
        search_func = search_funcs[type_]
    except IndexError: # pragma: no cover
        raise CraigslistValueError("unknown search type: {}".format(type_))

    search_gen = search_func(
        area, category, type_, cache, cachedir, executor, get, **kwargs)

    if get_detailed_posts:
        ret_gen = get_posts(extract_post_urls(search_gen), executor, get)
    else:
        ret_gen = search_gen

    return ret_gen

from craigslist._search.jsonsearch import jsonsearch_async

async def search_async(
    area,
    category,
    type_="jsonsearch",
    get_detailed_posts=False,
    cache=True,
    cachedir=os.path.expanduser('~'),
    max_workers=None,
    get=asyncio_get,
    **kwargs):

    async def extract_post_urls(posts):
        async for post in posts:
            yield post.url

    search_funcs = {"jsonsearch": jsonsearch_async}

    try:
        search_func = search_funcs[type_]
    except IndexError: # pragma: no cover
        raise CraigslistValueError("unknown search type: {}".format(type_))

    search_gen = search_func(
        area, category, type_, cache, cachedir, get, **kwargs)

    if get_detailed_posts:
        ret_gen = get_posts_async(extract_post_urls(search_gen), get)
    else:
        ret_gen = search_gen

    async for post in ret_gen:
        yield post



#!/usr/bin/env python3

from setuptools import setup, find_packages

readme = open('README.rst').read()
version = exec(open('craigslist/version.py').read())

setup(
    name='craigslist',
    version=__version__,
    description='Python wrapper for craigslist.',
    long_description=readme,
    author='Al Johri',
    author_email='al.johri@gmail.com',
    url='https://github.com/AlJohri/craigslist',
    license='MIT',
    packages=find_packages(),
    package_data={'craigslist': ['data/*.json']},
    install_requires=['requests', 'cssselect', 'lxml', 'arrow'],
    setup_requires=['pytest-runner'],
    tests_require=['pytest'],
    entry_points={
        'console_scripts': [
            'craigslist=craigslist.cli:main'
        ]
    },
    classifiers=[
        'Programming Language :: Python :: 3 :: Only',
        'Programming Language :: Python :: 3.6',
        'License :: OSI Approved :: MIT License',
    ]
)



#!/usr/bin/env python3

import sys
import json
import itertools
import argparse
import logging
import textwrap
from os import path
from craigslist import search
from craigslist.data import get_areas, get_categories, DATA_FOLDER
from craigslist.exceptions import CraigslistException
from craigslist._search import make_executor

class CustomArgumentParser(argparse.ArgumentParser):
    def error(self, message):
        """
        don't show error message if you just type "craigslist"
        """
        messages_to_mute = [
            "the following arguments are required: %s" % word for word in ('command',)
        ]
        if message not in messages_to_mute:
            sys.stderr.write('error: %s\n' % message)
        self.print_help()
        sys.exit(2)

def cli():
    global_description = """
    examples:
    craigslist search worcester apa
    craigslist search newyork tlg
    craigslist search washingtondc apa --postal 20071 --search_distance 1
    craigslist search newyork aap --postal 10023 --search_distance 1 --hasPic --availabilityMode within_30_days --limit 100
    craigslist search sfbay ccc --postal 94305 --search_distance 1 --limit 10
    craigslist search vancouver sss "shoes" --condition new like_new --hasPic --max_price 20 --limit 10
    craigslist search washingtondc jjj --is_telecommuting --is_internship
    craigslist list areas
    craigslist list categories
    """
    global_description = textwrap.dedent(global_description)
    formatter_class = lambda prog: argparse.RawDescriptionHelpFormatter(prog, max_help_position=32)

    parser = CustomArgumentParser(
        prog='craigslist',
        description=global_description,
        formatter_class=formatter_class)
    subparsers = parser.add_subparsers(dest='command', parser_class=CustomArgumentParser)
    subparsers.required = True

    shared_parser = argparse.ArgumentParser(add_help=False)
    shared_parser.add_argument('--verbose', default=False, action='store_true')

    with open(path.join(DATA_FOLDER, 'arguments.json')) as f:
        search_arguments = {x['dest']:x for x in json.load(f)}

    def create_search_parser(parent_subparsers, shared_parsers=[]):

        def cli_search(args):
            # TODO: make this a whitelist not a blacklist for starters
            filter_out_params = [
                'verbose', 'command', 'area', 'category', 'detail', 'executor_class', 'method', 'max_workers',
                'limit', 'cache', 'func']
            params = {k:v for k,v in vars(args).items() if v and k not in filter_out_params}
            logging.info('querying with parameters: {}'.format(params))

            # subclass ArgumentParser to make this happen automatically
            # it seems to stop using the `choices` parameter if nargs is defined
            for k,v in params.items():
                if k in search_arguments and\
                    search_arguments[k].get('nargs') == '*' and\
                    search_arguments[k].get('choices') is not None and\
                    isinstance(v, list):

                    mapping = search_arguments[k].get('choices')
                    params[k] = [mapping[x] for x in v]
            executor = make_executor(args.executor_class, args.max_workers)
            posts = itertools.islice(
                search(
                    args.area,
                    args.category,
                    get_detailed_posts=args.detail,
                    executor=executor,
                    type_=args.method,
                    **params), 0, args.limit)
            try:
                for post in posts:
                    print(json.dumps(post._asdict()))
                # TODO: this is wrong, if I want to limit, I need to do it
                # in some other way than by using islice
                # maybe I can using the signal library to send a signal?
                # https://docs.python.org/3/library/signal.html
                # executor.shutdown(wait=False)
                # sys.exit()
            except CraigslistException as e:
                print(e, file=sys.stderr)
                sys.exit()

        parser = parent_subparsers.add_parser(
            'search',
            usage='%(prog)s area category [options]',
            description=global_description,
            formatter_class=formatter_class,
            help='search',
            parents=shared_parsers)

        parser.add_argument('area')
        parser.add_argument('category')
        parser.add_argument('query', nargs='?', default=None)

        for dest, argument in search_arguments.items():
            x = {k:v for k,v in argument.items() if v is not None}
            parser.add_argument("--" + dest, **x)

        parser.add_argument('--limit', type=int)
        parser.add_argument('--detail', action="store_true")
        parser.add_argument('--executor_class', default='concurrent.futures.ThreadPoolExecutor')
        parser.add_argument('--method', default='jsonsearch')
        parser.add_argument('--max_workers', default=10)
        parser.add_argument('--cachedir', help='Cache directory. Defaults to ~/.craigslist')
        parser.add_argument('--nocache', action="store_false", dest='cache', default=True)
        parser.add_argument('--debug', dest='executor_class', action='store_const', const='craigslist.io.DebugExecutor')
        parser.set_defaults(func=cli_search)

    def create_list_parser(parent_subparsers, shared_parsers=[]):

        def cli_list(args):
            if args.entity == "areas":
                areas = get_areas()
                for hostname, area in areas.items():
                    print(json.dumps(area))
            elif args.entity == "categories":
                categories = get_categories()
                for abbreviation, category in categories.items():
                    print(json.dumps(category))
            else:
                raise Exception("don't know how to list {}".format(args.entity))

        parser = parent_subparsers.add_parser(
            'list',
            description=global_description,
            formatter_class=formatter_class,
            help='list',
            parents=shared_parsers)

        parser.add_argument('entity', metavar='entity',
            choices=['areas', 'categories'],
            help="Choices are: 'areas', 'categories'")
        parser.set_defaults(func=cli_list)

    create_search_parser(subparsers, shared_parsers=[shared_parser])
    create_list_parser(subparsers, shared_parsers=[shared_parser])

    args = parser.parse_args()

    if args.verbose:
        logging.basicConfig(level=logging.DEBUG, format=
            "[%(name)s | Thread: %(thread)d %(threadName)s | "
            "Process: %(process)d %(processName)s] %(asctime)s %(message)s")
        logging.getLogger('requests').setLevel(logging.WARNING)
        logging.getLogger('urllib3').setLevel(logging.WARNING)

    return args

# TODO modularize this more to make it testable
def main():
    args = cli()
    args.func(args)

if __name__ == '__main__':
    main()
